// DGT_BezierCam.js
//
// syntax:
// BCAM (START) (P1) (P2) (END) length mode
//         |     |    |     |     |     L____ can be left out, will be Linear by default (case sensitive), refer to the 'animation types' below
//         |     |    |     |     |
//         |     |    |     |     L__________ lower is faster, defines the number of frames that the camera path should take
//         |     |    |     |
//         L_____L____L_____L________________ each of these points define points in the bezier curve and follow the node syntax found below
//
// node syntax:
//   PLAYER:
//     refers to the current x/y position of the player, usually equal to CUR, but not always
//
//   CUR:
//     refers to the current position of the camera, most useful as the starting node, but could be used as any of the four nodes
//
//   X00Y00:
//     refers to a specific x/y coordinate position
//
//   E00:
//     refers to the position of a particular event with the specified id
//
//   XV00YV00:
//     uses the contents of the specified variables as x and y coordinates
//
//   EV00:
//     uses the contents of the specified variable as the id of an event to use the position of
//
// animation types:
//  QuadIn      QuadOut      QuadInOut
//  CubicIn     CubicOut     CubicInOut
//  QuartIn     QuartOut     QuartInOut
//  QuintIn     QuintOut     QuintInOut
//  SineIn      SineOut      SineInOut
//  ExpoIn      ExpoOut      ExpoInOut
//  CircIn      CircOut      CircInOut
//  ElasticIn   ElasticOut   ElasticInOut
//  BackIn      BackOut      BackInOut
//  BounceIn    BounceOut    BounceInOut
//
//  Linear (default)
//
// Versions:
//   1.0.0 - initial version

var DGT = DGT || {}
DGT.BCAM = DGT.BCAM || {}

{
    DGT.BCAM.old_fn_store = DGT.BCAM.old_fn_store || new Map()
    let _alias = (originalStorage, baseClass, funcName, newFunc) => {

        let storedBase = originalStorage.get(baseClass) || {} // ensure default
        originalStorage.set(baseClass, storedBase)
        storedBase[funcName] = baseClass[funcName] || (() => { }) // store old function
        baseClass[funcName] = function (...args) {
            return newFunc.call(this, storedBase[funcName], ...args) // call new function, passing the original function as the first argument
        }
    }
    let alias = _alias.bind(null, DGT.BCAM.old_fn_store)

    alias(Game_Interpreter.prototype, 'pluginCommand', function (original, command, args) {
        if (command.toLowerCase() !== "bcam") { return original.call(this, command, args) }
        let targets = args.slice(0, 4).map(resolveEvent)
        let length = parseInt(args[4])
        let mode = TDDP_AnimationCurves.easingFunctions[args[5]]
        if (targets.includes(null) || !length) {
            return console.error("INVALID BCAM PLUGIN COMMAND SYNTAX")
        }
        $gameMap.setBcamTarget(targets, length, mode)
    })

    const X_OFFSET = 9.5
    const Y_OFFSET = 6.625 // could not figure out how to derive this from base engine functions so HARDCODED VALUES IT IS

    function constructPsuedoEvent(x, y) {
        // x = x - X_OFFSET
        // y = y - Y_OFFSET // whY
        return ({
            x: x,
            y: y,
            _realX: x,
            _realY: y,
            screenX: Game_CharacterBase.prototype.screenX,
            screenY: function () {
                let th = $gameMap.tileHeight();
                return Math.round(this.scrolledY() * th + th);
            },
            scrolledX: Game_CharacterBase.prototype.scrolledX,
            scrolledY: Game_CharacterBase.prototype.scrolledY
        })
    }

    function resolveValue(val) {
        if (val[0] === 'v') {
            let varid = Number(val[0].replace('v', ''))
            return $gameVariables.value(varid)
        } else {
            return Number(val)
        }
    }

    function resolveEvent(loc) {
        loc = loc.toLowerCase()
        if (loc === "cur") {
            let x = $gameMap.displayX() + X_OFFSET
            let y = $gameMap.displayY() + Y_OFFSET
            return constructPsuedoEvent(x, y)
        }
        if (loc === "player") {
            return constructPsuedoEvent($gamePlayer.x, $gamePlayer.y)
        }
        if (loc.match(/x(v?[0-9.]+)y(v?[0-9.]+)/)) {
            let x = resolveValue(RegExp.$1)
            let y = resolveValue(RegExp.$2)
            return constructPsuedoEvent(x, y)
        }
        if (loc.match(/e(v?[0-9.]+)/)) {
            let eventid = resolveValue(RegExp.$1)
            return $gameMap.event(eventid)
        }
        return null
    }
    DGT.BCAM.resolveEvent = resolveEvent
    alias(Game_Map.prototype, 'initBcam', function () {
        this.bcam = {
            norm: true,
            targets: [],
            length: 500,
            frames: 0,
            lastx: 0,
            lasty: 0,
            mode: TDDP_AnimationCurves.easingFunctions.Linear
        }
    })
    alias(Game_Map.prototype, 'setup', function (original, ...args) {
    	this.initBcam()
        original.apply(this, args)
    })

    alias(Game_Map.prototype, 'setBcamTarget', function (_, nodes, length, mode) {
    	if (!this.bcam) { this.initBcam() }
        this.bcam.targets = nodes
        this.bcam.length = length
        this.bcam.norm = false
        this.bcam.frames = 0
        this.bcam.lastx = 0
        this.bcam.lasty = 0
        this.bcam.mode = mode || TDDP_AnimationCurves.easingFunctions.Linear
    })

    alias(Game_Map.prototype, 'updateScroll', function (original) {
    	if (!this.bcam) { this.initBcam() }
        if (this.bcam.norm) { return original.call(this) }
        
        var cw = (Graphics.boxWidth / 2);
        var ch = (Graphics.boxHeight / 2);
        if (this.bcam.frames === 0) {
            let cx = $gameMap.displayX() + X_OFFSET
            let cy = $gameMap.displayY() + Y_OFFSET
            let fx = this.bcam.targets[0].x
            let fy = this.bcam.targets[0].y
            let dx = fx - cx
            let dy = fy - cy
            if (dy > ch) { this.scrollUp(dy) } else if (dy < ch) { this.scrollDown(dy) };
            if (dx > cw) { this.scrollLeft(dx) } else if (dx < cw) { this.scrollRight(dx) };
        }
        
        this._scrollRest = 0;
        this.bcam.frames++
        
        let x0 = this.bcam.targets[0].x
        let y0 = this.bcam.targets[0].y
        let x1 = this.bcam.targets[1].x
        let y1 = this.bcam.targets[1].y
        let x2 = this.bcam.targets[2].x
        let y2 = this.bcam.targets[2].y
        let x3 = this.bcam.targets[3].x
        let y3 = this.bcam.targets[3].y

        let L = TDDP_AnimationCurves.easingFunctions.Linear
        let l0x0 = L(this.bcam.frames, x0, x1 - x0, this.bcam.length)
        let l0y0 = L(this.bcam.frames, y0, y1 - y0, this.bcam.length)
        let l0x1 = L(this.bcam.frames, x1, x2 - x1, this.bcam.length)
        let l0y1 = L(this.bcam.frames, y1, y2 - y1, this.bcam.length)
        let l0x2 = L(this.bcam.frames, x2, x3 - x2, this.bcam.length)
        let l0y2 = L(this.bcam.frames, y2, y3 - y2, this.bcam.length)

        let l1x0 = L(this.bcam.frames, l0x0, l0x1 - l0x0, this.bcam.length)
        let l1y0 = L(this.bcam.frames, l0y0, l0y1 - l0y0, this.bcam.length)
        let l1x1 = L(this.bcam.frames, l0x1, l0x2 - l0x1, this.bcam.length)
        let l1y1 = L(this.bcam.frames, l0y1, l0y2 - l0y1, this.bcam.length)

        let C = this.bcam.mode
        let fx = C(this.bcam.frames, l1x0, l1x1 - l1x0, this.bcam.length)
        let fy = C(this.bcam.frames, l1y0, l1y1 - l1y0, this.bcam.length)

        let dx = fx - this.lastx
        let dy = fy - this.lasty

        this.lastx = fx
        this.lasty = fy

        if (dy > ch) { this.scrollUp(dy) } else if (dy < ch) { this.scrollDown(dy) };

        if (dx > cw) { this.scrollLeft(dx) } else if (dx < cw) { this.scrollRight(dx) };

        if (this.bcam.frames === this.bcam.length) { this.bcam.norm = true }
    })

    // alias(Game_Variables.prototype, 'value', function(original, id) {
    //     let val = original.call(this, id)
    //     if (parseInt(val)) {
    //         return Math.randomInt(14)
    //     }
    //     return val
    // })
}